@@ -6,9 +6,11 @@ import dateutil.tz
 import pytz
 import datetime
 import xml.dom.minidom
+from sql import Table, Column
+
 from trytond.model import Model, ModelSQL, ModelView, fields
 from trytond.tools import reduce_ids
-from trytond.backend import TableHandler
+from trytond import backend
 from trytond.pyson import If, Bool, Eval
 from trytond.transaction import Transaction
 from trytond.cache import Cache
@@ -60,8 +62,8 @@ class Calendar(ModelSQL, ModelView):
         return calendars
 
     @classmethod
-    def write(cls, calendars, vals):
-        super(Calendar, cls).write(calendars, vals)
+    def write(cls, calendars, values, *args):
+        super(Calendar, cls).write(calendars, values, *args)
         # Restart the cache for get_name
         cls._get_name_cache.clear()
 
@@ -521,12 +523,6 @@ class Event(ModelSQL, ModelView):
     def default_transp():
         return 'opaque'
 
-    @staticmethod
-    def default_timezone():
-        User = Pool().get('res.user')
-        user = User(Transaction().user)
-        return user.timezone
-
     @staticmethod
     def timezones():
         return [(x, x) for x in pytz.common_timezones] + [('', '')]
@@ -629,99 +625,112 @@ class Event(ModelSQL, ModelView):
             'location': self.location.id if self.location else None,
             'status': self.status,
             'organizer': self.organizer,
-            'rdates': [('delete_all',)] + ('create', [rdate._date2update()
-                for rdate in self.rdates]),
-            'exdates': [('delete_all',)] + ('create', [exdate._date2update()
-                for exdate in self.exdates]),
-            'rrules': [('delete_all',)] + ('create', [rrule._date2update()
-                for rrule in self.rrules]),
-            'exrules': [('delete_all',)] + ('create', [exrule._date2update()
-                for exrule in self.exrules]),
+            'rdates': [('delete', [r.id for r in self.rdates])]
+            + [('create', [rdate._date2update()
+                        for rdate in self.rdates])],
+            'exdates': [('delete', [r.id for r in self.exdates])]
+            + [('create', [exdate._date2update()
+                        for exdate in self.exdates])],
+            'rrules': [('delete', [r.id for r in self.rrules])]
+            + [('create', [rrule._date2update()
+                        for rrule in self.rrules])],
+            'exrules': [('delete', [r.id for r in self.exrules])]
+            + [('create', [exrule._date2update()
+                        for exrule in self.exrules])],
             }
 
     @classmethod
-    def write(cls, events, values):
+    def write(cls, *args):
         pool = Pool()
         Calendar = pool.get('calendar.calendar')
         Collection = pool.get('webdav.collection')
         cursor = Transaction().cursor
 
-        values = values.copy()
-        if 'sequence' in values:
-            del values['sequence']
+        actions = iter(args)
+        args = []
+        for events, values in zip(actions, actions):
+            values = values.copy()
+            if 'sequence' in values:
+                del values['sequence']
+            args.extend((events, values))
 
-        super(Event, cls).write(events, values)
+        super(Event, cls).write(*args)
 
+        table = cls.__table__()
         for i in range(0, len(events), cursor.IN_MAX):
             sub_ids = map(int, events[i:i + cursor.IN_MAX])
-            red_sql, red_ids = reduce_ids('id', sub_ids)
-            cursor.execute('UPDATE "' + cls._table + '" '
-                'SET sequence = sequence + 1 '
-                'WHERE ' + red_sql, red_ids)
+            red_sql = reduce_ids(table.id, sub_ids)
+            cursor.execute(*table.update(
+                    columns=[table.sequence],
+                    values=[table.sequence + 1],
+                    where=red_sql))
 
-        if not values:
-            return
-        for event in events:
-            if (event.calendar.owner
-                    and (event.organizer == event.calendar.owner.email
-                        or (event.parent
-                            and event.parent.organizer
-                            == event.calendar.owner.email))):
-                if event.organizer == event.calendar.owner.email:
-                    attendee_emails = [x.email for x in event.attendees
-                            if x.status != 'declined'
-                            and x.email != event.organizer]
-                else:
-                    attendee_emails = [x.email for x in event.parent.attendees
+        actions = iter(args)
+        for events, values in zip(actions, actions):
+            if not values:
+                continue
+            for event in events:
+                if (event.calendar.owner
+                        and (event.organizer == event.calendar.owner.email
+                            or (event.parent
+                                and event.parent.organizer
+                                == event.calendar.owner.email))):
+                    if event.organizer == event.calendar.owner.email:
+                        attendee_emails = [x.email for x in event.attendees
+                                if x.status != 'declined'
+                                and x.email != event.organizer]
+                    else:
+                        attendee_emails = [x.email
+                            for x in event.parent.attendees
                             if x.status != 'declined'
                             and x.email != event.parent.organizer]
-                with Transaction().set_user(0):
-                    events2 = cls.search([
-                            ('uuid', '=', event.uuid),
-                            ('id', '!=', event.id),
-                            ('recurrence', '=', event.recurrence),
-                            ])
-                    for event2 in events2[:]:
-                        if event2.calendar.owner.email in attendee_emails:
-                            attendee_emails.remove(
-                                    event2.calendar.owner.email)
-                        else:
-                            events2.remove(event2)
-                            cls.delete([event2])
-                    if events2:
-                        cls.write(events2, event._event2update())
-                if attendee_emails:
                     with Transaction().set_user(0):
-                        calendars = Calendar.search([
-                                ('owner.email', 'in', attendee_emails),
+                        events2 = cls.search([
+                                ('uuid', '=', event.uuid),
+                                ('id', '!=', event.id),
+                                ('recurrence', '=', event.recurrence),
                                 ])
-                        if not event.recurrence:
-                            for calendar in calendars:
-                                new_event, = cls.copy([event], default={
-                                        'calendar': calendar.id,
-                                        'occurences': None,
-                                        'uuid': event.uuid,
-                                        })
-                                for occurence in event.occurences:
-                                    cls.copy([occurence], default={
-                                            'calendar': calendar.id,
-                                            'parent': new_event.id,
-                                            'uuid': occurence.uuid,
-                                            })
-                        else:
-                            parents = cls.search([
-                                    ('uuid', '=', event.uuid),
-                                    ('calendar.owner.email', 'in',
-                                        attendee_emails),
-                                    ('id', '!=', event.id),
-                                    ('recurrence', '=', None),
+                        for event2 in events2[:]:
+                            if event2.calendar.owner.email in attendee_emails:
+                                attendee_emails.remove(
+                                        event2.calendar.owner.email)
+                            else:
+                                events2.remove(event2)
+                                cls.delete([event2])
+                        if events2:
+                            cls.write(events2, event._event2update())
+                    if attendee_emails:
+                        with Transaction().set_user(0):
+                            calendars = Calendar.search([
+                                    ('owner.email', 'in', attendee_emails),
                                     ])
-                            for parent in parents:
-                                cls.copy([event], default={
-                                        'calendar': parent.calendar.id,
-                                        'parent': parent.id,
-                                        'uuid': event.uuid,
-                                        })
+                            if not event.recurrence:
+                                for calendar in calendars:
+                                    new_event, = cls.copy([event], default={
+                                            'calendar': calendar.id,
+                                            'occurences': None,
+                                            'uuid': event.uuid,
+                                            })
+                                    for occurence in event.occurences:
+                                        cls.copy([occurence], default={
+                                                'calendar': calendar.id,
+                                                'parent': new_event.id,
+                                                'uuid': occurence.uuid,
+                                                })
+                            else:
+                                parents = cls.search([
+                                        ('uuid', '=', event.uuid),
+                                        ('calendar.owner.email', 'in',
+                                            attendee_emails),
+                                        ('id', '!=', event.id),
+                                        ('recurrence', '=', None),
+                                        ])
+                                for parent in parents:
+                                    cls.copy([event], default={
+                                            'calendar': parent.calendar.id,
+                                            'parent': parent.id,
+                                            'uuid': event.uuid,
+                                            })
         # Restart the cache for event
         Collection._event_cache.clear()
 
@@ -802,7 +811,7 @@ class Event(ModelSQL, ModelView):
         Alarm = pool.get('calendar.event.alarm')
         Rdate = pool.get('calendar.event.rdate')
         Exdate = pool.get('calendar.event.exdate')
-        Rrule = pool.get('calendar.rrule')
+        Rrule = pool.get('calendar.event.rrule')
         Exrule = pool.get('calendar.event.exrule')
 
         vevents = []
@@ -872,6 +881,7 @@ class Event(ModelSQL, ModelView):
             res['status'] = vevent.status.value.lower()
         else:
             res['status'] = ''
+        res['categories'] = [('remove', [c.id for c in event.categories])]
         if hasattr(vevent, 'categories'):
             with Transaction().set_context(active_test=False):
                 categories = Category.search([
@@ -888,9 +898,7 @@ class Event(ModelSQL, ModelView):
                             })
             if to_create:
                 categories += Category.create(to_create)
-            res['categories'] = [('set', map(int, categories))]
-        else:
-            res['categories'] = [('unlink_all',)]
+            res['categories'] += [('add', map(int, categories))]
         if hasattr(vevent, 'class'):
             if getattr(vevent, 'class').value.lower() in \
                     dict(cls.classification.selection):
@@ -1048,15 +1056,9 @@ class Event(ModelSQL, ModelView):
         '''
         Return an iCalendar instance of vobject for event
         '''
-        pool = Pool()
-        User = pool.get('res.user')
-
-        user = User(Transaction().user)
         if self.timezone:
             tzevent = pytz.timezone(self.timezone)
             tzevent = dateutil.tz.gettz(self.timezone)
-        elif user.timezone:
-            tzevent = dateutil.tz.gettz(user.timezone)
         else:
             tzevent = tzlocal
 
@@ -1230,7 +1232,9 @@ class EventAlarm(AlarmMixin, ModelSQL, ModelView):
 
     @classmethod
     def __register__(cls, module_name):
+        TableHandler = backend.get('TableHandler')
         cursor = Transaction().cursor
+        sql_table = cls.__table__()
 
         super(EventAlarm, cls).__register__(module_name)
 
@@ -1238,10 +1242,11 @@ class EventAlarm(AlarmMixin, ModelSQL, ModelView):
 
         # Migration from 2.6: Remove inherits calendar.alarm
         if table.column_exist('calendar_alarm'):
-            cursor.execute('UPDATE "' + cls._table + '" AS e '
-                'SET valarm = (SELECT a.valarm '
-                    'FROM calendar_alarm AS a '
-                    'WHERE a.id = e.calendar_alarm)')
+            alarm = Table('calendar_alarm')
+            cursor.execute(*sql_table.update(
+                    columns=[sql_table.valarm],
+                    values=[alarm.select(alarm.valarm,
+                            where=alarm.id == sql_table.calendar_alarm)]))
             table.drop_column('calendar_alarm', True)
 
     @classmethod
@@ -1257,15 +1262,19 @@ class EventAlarm(AlarmMixin, ModelSQL, ModelView):
         return super(EventAlarm, cls).create(vlist)
 
     @classmethod
-    def write(cls, event_alarms, values):
+    def write(cls, *args):
         Event = Pool().get('calendar.event')
-        events = [x.event for x in event_alarms]
-        if values.get('event'):
-            events.append(Event(values['event']))
+
+        actions = iter(args)
+        events = []
+        for event_alarms, values in zip(actions, actions):
+            events += [x.event for x in event_alarms]
+            if values.get('event'):
+                events.append(Event(values['event']))
         if events:
             # Update write_date of event
             Event.write(events, {})
-        return super(EventAlarm, cls).write(event_alarms, values)
+        super(EventAlarm, cls).write(*args)
 
     @classmethod
     def delete(cls, event_alarms):
@@ -1352,7 +1361,9 @@ class EventAttendee(AttendeeMixin, ModelSQL, ModelView):
 
     @classmethod
     def __register__(cls, module_name):
+        TableHandler = backend.get('TableHandler')
         cursor = Transaction().cursor
+        sql_table = cls.__table__()
 
         super(EventAttendee, cls).__register__(module_name)
 
@@ -1360,13 +1371,14 @@ class EventAttendee(AttendeeMixin, ModelSQL, ModelView):
 
         # Migration from 2.6: Remove inherits calendar.attendee
         if table.column_exist('calendar_attendee'):
-            cursor.execute('UPDATE "' + cls._table + '" AS e '
-                'SET email = (SELECT a.email '
-                    'FROM calendar_attendee AS a '
-                    'WHERE a.id = e.calendar_attendee), '
-                'status = (SELECT a.status '
-                    'FROM calendar_attendee AS a '
-                    'WHERE a.id = e.calendar_attendee)')
+            attendee = Table('calendar_attendee')
+            cursor.execute(*sql_table.update(
+                    columns=[sql_table.email, sql_table.status],
+                    values=[attendee.select(attendee.email,
+                            where=attendee.id == sql_table.calendar_attendee),
+                        attendee.select(attendee.status,
+                            where=attendee.id == sql_table.calendar_attendee),
+                        ]))
             table.drop_column('calendar_attendee', True)
 
     @classmethod
@@ -1379,7 +1391,7 @@ class EventAttendee(AttendeeMixin, ModelSQL, ModelView):
                 to_write.append(values['event'])
 
         if to_write:
-            Event.write([Event.browse(to_write)], {})
+            Event.write(Event.browse(to_write), {})
         event_attendees = super(EventAttendee, cls).create(vlist)
         for event_attendee in event_attendees:
             event = event_attendee.event
@@ -1410,21 +1422,28 @@ class EventAttendee(AttendeeMixin, ModelSQL, ModelView):
         return event_attendees
 
     @classmethod
-    def write(cls, event_attendees, values):
+    def write(cls, *args):
         Event = Pool().get('calendar.event')
-        events = [x.event for x in event_attendees]
-        if values.get('event'):
-            events.append(Event(values['event']))
+
+        actions = iter(args)
+        args = []
+        events = []
+        for event_attendees, values in zip(actions, actions):
+            events += [x.event for x in event_attendees]
+            if values.get('event'):
+                events.append(Event(values['event']))
+            if 'email' in values:
+                values = values.copy()
+                del values['email']
+            args.extend((event_attendees, values))
+
         if events:
             # Update write_date of event
             Event.write(events, {})
 
-        if 'email' in values:
-            values = values.copy()
-            del values['email']
+        super(EventAttendee, cls).write(*args)
 
-        super(EventAttendee, cls).write(event_attendees, values)
-        for event_attendee in event_attendees:
+        for event_attendee in sum(args[::2], []):
             event = event_attendee.event
             if (event.calendar.owner
                     and (event.organizer == event.calendar.owner.email
@@ -1559,7 +1578,9 @@ class EventRDate(DateMixin, ModelSQL, ModelView):
 
     @classmethod
     def __register__(cls, module_name):
+        TableHandler = backend.get('TableHandler')
         cursor = Transaction().cursor
+        sql_table = cls.__table__()
         # Migration from 1.4: calendar_rdate renamed to calendar_date
         table = TableHandler(cursor, cls, module_name)
         old_column = 'calendar_rdate'
@@ -1572,13 +1593,13 @@ class EventRDate(DateMixin, ModelSQL, ModelView):
 
         # Migration from 2.6: Remove inherits calendar.date
         if table.column_exist('calendar_date'):
-            cursor.execute('UPDATE "' + cls._table + '" AS e '
-                'SET date = (SELECT a.date '
-                    'FROM calendar_date AS a '
-                    'WHERE a.id = e.calendar_date), '
-                'datetime = (SELECT a.datetime '
-                    'FROM calendar_date AS a '
-                    'WHERE a.id = e.calendar_date)')
+            date = Table('calendar_date')
+            cursor.execute(*sql_table.update(
+                    columns=[sql_table.date, sql_table.datetime],
+                    values=[date.select(date.date,
+                            where=date.id == sql_table.calendar_date),
+                        date.select(date.datetime,
+                            where=date.id == sql_table.calendar_date)]))
             table.drop_column('calendar_date', True)
 
     @classmethod
@@ -1594,15 +1615,19 @@ class EventRDate(DateMixin, ModelSQL, ModelView):
         return super(EventRDate, cls).create(vlist)
 
     @classmethod
-    def write(cls, event_rdates, values):
+    def write(cls, *args):
         Event = Pool().get('calendar.event')
-        events = [x.event for x in event_rdates]
-        if values.get('event'):
-            events.append(Event(values['event']))
+
+        actions = iter(args)
+        events = []
+        for event_rdates, values in zip(actions, actions):
+            events += [x.event for x in event_rdates]
+            if values.get('event'):
+                events.append(Event(values['event']))
         if events:
             # Update write_date of event
             Event.write(events, {})
-        super(EventRDate, cls).write(event_rdates, values)
+        super(EventRDate, cls).write(*args)
 
     @classmethod
     def delete(cls, event_rdates):
@@ -1647,6 +1672,7 @@ class RRuleMixin(Model):
     bymonth = fields.Char('By Month')
     bysetpos = fields.Char('By Position')
     wkst = fields.Selection([
+        (None, ''),
         ('su', 'Sunday'),
         ('mo', 'Monday'),
         ('tu', 'Tuesday'),
@@ -1677,9 +1703,10 @@ class RRuleMixin(Model):
                     'rule "%s"'),
                 'invalid_byweekno': ('Invalid "By Week Number" in recurrence '
                     'rule "%s"'),
-                'invalid_bymonth': 'Invalid "By Month" in recurrence rule "%s"',
-                'invalid_bysetpos': ('Invalid "By Position" in recurrence rule '
-                    '"%s"'),
+                'invalid_bymonth': (
+                    'Invalid "By Month" in recurrence rule "%s"'),
+                'invalid_bysetpos': (
+                    'Invalid "By Position" in recurrence rule "%s"'),
                 })
 
     @classmethod
@@ -1761,7 +1788,8 @@ class RRuleMixin(Model):
                 except Exception:
                     yeardaynum = -1000
                 if not (abs(yeardaynum) >= 1 and abs(yeardaynum) <= 366):
-                    self.raise_user_error('invalid_byyearday', (self.rec_name,))
+                    self.raise_user_error('invalid_byyearday',
+                        (self.rec_name,))
 
     def check_byweekno(self):
         if self.byweekno:
@@ -1865,7 +1893,9 @@ class EventRRule(RRuleMixin, ModelSQL, ModelView):
 
     @classmethod
     def __register__(cls, module_name):
+        TableHandler = backend.get('TableHandler')
         cursor = Transaction().cursor
+        sql_table = cls.__table__()
 
         super(EventRRule, cls).__register__(module_name)
 
@@ -1873,13 +1903,13 @@ class EventRRule(RRuleMixin, ModelSQL, ModelView):
 
         # Migration from 2.6: Remove inherits calendar.rrule
         if table.column_exist('calendar_rrule'):
+            rrule = Table('calendar_rrule')
             for field in (f for f in dir(RRuleMixin)
                     if isinstance(f, fields.Field)):
-                cursor.execute(('UPDATE "' + cls._table + '" AS e '
-                        'SET "%(field)s" = (SELECT a."%(field)s" '
-                            'FROM calendar_rrule AS r '
-                            'WHERE r.id = e.calendar_rrule)')
-                    % {'field': field})
+                cursor.execute(*sql_table.update(
+                        columns=[Column(sql_table, field)],
+                        values=[rrule.select(Column(rrule, field),
+                                where=rrule.id == sql_table.calendar_rrule)]))
             table.drop_column('calendar_rrule', True)
 
     @classmethod
@@ -1895,15 +1925,19 @@ class EventRRule(RRuleMixin, ModelSQL, ModelView):
         return super(EventRRule, cls).create(vlist)
 
     @classmethod
-    def write(cls, event_rrules, values):
+    def write(cls, *args):
         Event = Pool().get('calendar.event')
-        events = [x.event for x in event_rrules]
-        if values.get('event'):
-            events.append(Event(values['event']))
+
+        actions = iter(args)
+        events = []
+        for event_rrules, values in zip(actions, actions):
+            events += [x.event for x in event_rrules]
+            if values.get('event'):
+                events.append(Event(values['event']))
         if events:
             # Update write_date of event
             Event.write(events, {})
-        super(EventRRule, cls).write(event_rrules, values)
+        super(EventRRule, cls).write(*args)
 
     @classmethod
     def delete(cls, event_rrules):
